import pytest
import os
import numpy as np
import corgidrp
from corgidrp.mocks import create_default_L3_headers
from corgidrp.mocks import create_flux_image
from corgidrp.data import Image, Dataset, FluxcalFactor
import corgidrp.fluxcal as fluxcal
import corgidrp.l4_to_tda as l4_to_tda
from astropy.modeling.models import BlackBody
import astropy.units as u

data = np.ones([1024,1024]) * 2 
err = np.ones([1,1024,1024]) * 0.5
prhd, exthd = create_default_L3_headers()
exthd["CFAMNAME"] = '3C'
exthd["FPAMNAME"] = 'ND475'
prhd["TARGET"] = 'VEGA'
image1 = Image(data,pri_hdr = prhd, ext_hdr = exthd, err = err)
image2 = image1.copy()
image1.filename = "test1_L4_.fits"
image2.filename = "test2_L4_.fits"
dataset=Dataset([image1, image2])
calspec_filepath = os.path.join(os.path.dirname(__file__), "test_data", "alpha_lyr_stis_011.fits")

def test_get_filter_name():
    """
    test that the correct filter curve file is selected
    """
    global wave
    global transmission
    filepath = fluxcal.get_filter_name(image1)
    assert filepath.split("/")[-1] == 'transmission_ID-21_3C_v0.csv'
    
    wave, transmission = fluxcal.read_filter_curve(filepath)
    
    assert np.any(wave>=7130)
    assert np.any(transmission < 1.)
    
    #test a wrong filter name
    image3 = image1.copy()
    image3.ext_hdr["CFAMNAME"] = '5C'
    dataset2 = Dataset([image3, image3])
    with pytest.raises(ValueError):
        filepath = fluxcal.get_filter_name(image3)
        pass
    

def test_flux_calc():
    """
    test that the calspec data is read correctly
    """
    global band_flux
    calspec_flux = fluxcal.read_cal_spec(calspec_filepath, wave)
    assert calspec_flux[0] == pytest.approx(1.6121e-09, 1e-15) 
    
    band_flux = fluxcal.calculate_band_flux(transmission, calspec_flux, wave)
    eff_lambda = fluxcal.calculate_effective_lambda(transmission, calspec_flux, wave)
    assert eff_lambda == pytest.approx((wave[0]+wave[-1])/2., 3)
    
def test_colorcor():
    """
    test that the pivot reference wavelengths is close to the center of the bandpass
    and the color correction is calculated correctly
    """
    
    lambda_piv = fluxcal.calculate_pivot_lambda(transmission, wave)
    assert lambda_piv == pytest.approx((wave[0]+wave[-1])/2., 0.3)
    
    calspec_flux = fluxcal.read_cal_spec(calspec_filepath, wave)
    ## BB of an O5 star
    bbscale = 1.e-21 * u.erg/(u.s * u.cm**2 * u.AA * u.steradian)
    flux_source = BlackBody(scale = bbscale, temperature=54000.0 * u.K)
    K_bb = fluxcal.compute_color_cor(transmission, wave, calspec_flux, lambda_piv, flux_source(wave))
    assert K_bb == pytest.approx(1., 0.01)
    
    flux_source = BlackBody(scale = bbscale, temperature=100. * u.K)
    K_bb = fluxcal.compute_color_cor(transmission, wave, calspec_flux, lambda_piv, flux_source(wave))
    assert K_bb > 2#weakest star to be detected
    # sanity check
    K = fluxcal.compute_color_cor(transmission, wave, calspec_flux, lambda_piv, calspec_flux)
    assert K == 1 

    # test the corresponding pipeline step
    output_dataset = l4_to_tda.determine_color_cor(dataset, calspec_filepath, calspec_filepath)
    assert output_dataset[0].ext_hdr['LAM_REF'] == lambda_piv
    assert output_dataset[0].ext_hdr['COL_COR'] == K
    # test it with star names
    calspec_name = 'Vega'
    source_name = 'TYC 4424-1286-1'
    output_dataset = l4_to_tda.determine_color_cor(dataset, calspec_name, source_name)
    assert output_dataset[0].ext_hdr['LAM_REF'] == lambda_piv
    assert output_dataset[0].ext_hdr['COL_COR'] == pytest.approx(1,1e-2) 
    
def test_calspec_download():
    """
    test the download of a calspec fits file
    """
    filepath = fluxcal.get_calspec_file('Vega')
    assert os.path.exists(filepath)
    os.remove(filepath)
    filepath = fluxcal.get_calspec_file('TYC 4424-1286-1')
    assert os.path.exists(filepath)
    os.remove(filepath)
    
    with pytest.raises(ValueError):
        filepath = fluxcal.get_calspec_file('Todesstern')

def test_app_mag():
    """
    test the calculation of the apparent Vega magnitude
    """
    # test the corresponding pipeline step
    # sanity check
    output_dataset = l4_to_tda.determine_app_mag(dataset, 'Vega')
    assert output_dataset[0].ext_hdr['APP_MAG'] == 0.
    output_dataset = l4_to_tda.determine_app_mag(dataset, calspec_filepath)
    assert output_dataset[0].ext_hdr['APP_MAG'] == pytest.approx(0.0, 0.03) 
    output_dataset = l4_to_tda.determine_app_mag(dataset, calspec_filepath, scale_factor = 0.5)
    assert output_dataset[0].ext_hdr['APP_MAG'] == pytest.approx(0.+-2.5*np.log10(0.5), 0.03)
    output_dataset = l4_to_tda.determine_app_mag(dataset, '109 Vir')
    assert output_dataset[0].ext_hdr['APP_MAG'] == pytest.approx(3.72, 0.05)

def test_fluxcal_file():
    """ 
    Generate a mock fluxcal factor cal object and test the content and functionality.
    """
    fluxcal_factor = np.array([[2e-12]])
    fluxcal_factor_error = np.array([[[1e-14]]])
    fluxcal_fac = FluxcalFactor(fluxcal_factor, err = fluxcal_factor_error, pri_hdr = prhd, ext_hdr = exthd, input_dataset = dataset)
    assert fluxcal_fac.filter == '3C'
    assert fluxcal_fac.fluxcal_fac == fluxcal_factor[0,0]
    assert fluxcal_fac.fluxcal_err == fluxcal_factor_error[0,0,0]
    assert(fluxcal_fac.filename.split(".")[0] == "test2_ABF_CAL")
    
    calibdir = os.path.join(os.path.dirname(__file__), "testcalib")
    filename = fluxcal_fac.filename
    if not os.path.exists(calibdir):
        os.mkdir(calibdir)
    fluxcal_fac.save(filedir=calibdir, filename=filename)        
        
    fluxcal_filepath = os.path.join(calibdir, filename)

    fluxcal_fac_file = FluxcalFactor(fluxcal_filepath)
    assert fluxcal_fac_file.filter == '3C'
    assert fluxcal_fac_file.fluxcal_fac == fluxcal_factor[0,0]
    assert fluxcal_fac_file.fluxcal_err == fluxcal_factor_error[0,0,0]
    # JM: I moved this out of the fluxcal class and into fluxcal.py because, depending on the method you use to 
    # make the fluxcal factor, the BUNIT will vary. Doing a mock without running fluxcal methods won't update BUNIT
    #assert fluxcal_fac_file.ext_hdr["BUNIT"] == 'erg/(s * cm^2 * AA)/(electron/s)'

def test_abs_fluxcal():
    """ 
    Generate a simulated image and test the flux calibration computation.
    
    """
    # create a simulated image with source guesses and true positions
    # check that the simulated image folder exists and create if not
    datadir = os.path.join(os.path.dirname(__file__), "test_data", "sim_fluxcal")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    
    fwhm = 3
    cal_factor = band_flux/200
    #create a simulated mock image with a central point source + noise that has a flux band_flux 
    #and a flux calibration factor band_flux/200
    #that results in a total extracted count of 200 photo electrons
    flux_image = create_flux_image(band_flux, fwhm, cal_factor, filter='3C', target_name='Vega', fsm_x=0.0, 
                      fsm_y=0.0, exptime=1.0, filedir=datadir, platescale=21.8, background=0,
                      add_gauss_noise=True, noise_scale=1., file_save=True)
    assert type(flux_image) == Image
    sigma = fwhm/(2.*np.sqrt(2*np.log(2)))
    radius = 3.* sigma
    
    #Test the aperture photometry
    #The error of one pixel is 1, so the error of the aperture sum should be: 
    error_sum = np.sqrt(np.pi * radius * radius)
    [flux_el_ap, flux_err_ap] = fluxcal.aper_phot(flux_image, radius, frac_enc_energy=0.997, method='subpixel', subpixels=5,
              background_sub=False, r_in=5, r_out=10, centering_method='xy', centroid_roi_radius=5)
    #200 is the input count of photo electrons
    assert flux_el_ap == pytest.approx(200, rel = 0.05)
    assert flux_err_ap == pytest.approx(error_sum, rel = 0.05)
    
    
    #Test the 2D Gaussian fit photometry
    #The error of one pixel is 1, so the error of a circular aperture with radius of 2 sigma should be about:
    error_gauss = np.sqrt(np.pi * 4 * sigma * sigma)
    flux_el_gauss, flux_err_gauss = fluxcal.phot_by_gauss2d_fit(flux_image, fwhm, fit_shape = 41)
    assert flux_el_gauss == pytest.approx(200, rel = 0.05)
    assert flux_err_gauss == pytest.approx(error_gauss, rel = 0.05)
    
    flux_el_gauss, flux_err_gauss = fluxcal.phot_by_gauss2d_fit(flux_image, fwhm)
    assert flux_el_gauss == pytest.approx(200, rel = 0.05)
    assert flux_err_gauss == pytest.approx(error_gauss, rel = 0.05)
    #test the generation of the flux cal factors cal file
    dataset = Dataset([flux_image])
    fluxcal_factor = fluxcal.calibrate_fluxcal_aper(dataset, flux_or_irr = 'flux', phot_kwargs=None)
    assert fluxcal_factor.filter == '3C'
    #band_flux/200 was the input calibration factor cal_factor of the simulated mock image
    assert fluxcal_factor.fluxcal_fac == pytest.approx(cal_factor, rel = 0.05)
    #divisive error propagation of the aperture phot error
    err_fluxcal_ap = band_flux/flux_el_ap**2*flux_err_ap
    assert fluxcal_factor.fluxcal_err == pytest.approx(err_fluxcal_ap)
    # TO DO: add this test back in when filename conventions are settled
    #assert fluxcal_factor.filename == 'mock_flux_image_Vega_0.0_0.0_FluxcalFactor_3C_ND475.fits'
    fluxcal_factor_gauss = fluxcal.calibrate_fluxcal_gauss2d(dataset, flux_or_irr = 'flux', phot_kwargs=None)
    assert fluxcal_factor_gauss.filter == '3C'
    assert fluxcal_factor_gauss.fluxcal_fac == pytest.approx(cal_factor,rel = 0.05)
    #divisive error propagation of the 2D Gaussian fit phot error
    err_fluxcal_gauss = band_flux/flux_el_gauss**2*flux_err_gauss
    assert fluxcal_factor_gauss.fluxcal_err == pytest.approx(err_fluxcal_gauss)
    
    #test the flux conversion computation.
    old_ind = corgidrp.track_individual_errors
    corgidrp.track_individual_errors = True
    flux_dataset = Dataset([flux_image, flux_image])
    output_dataset = l4_to_tda.convert_to_flux(flux_dataset, fluxcal_factor)
    assert len(output_dataset) == 2
    assert output_dataset[0].ext_hdr['BUNIT'] == "erg/(s*cm^2*AA)"
    assert output_dataset[0].ext_hdr['FLUXFAC'] == fluxcal_factor.fluxcal_fac
    assert "fluxcal_factor" in str(output_dataset[0].ext_hdr['HISTORY'])
    
    #this is the amplitude of the simulated 2D Gaussian mock image in the center
    #it should be close to the value in the output image center
    el_cen = flux_dataset[0].data[512,512]
    amplitude = el_cen * fluxcal_factor.fluxcal_fac
    data_cen = output_dataset[0].data[512, 512]
    assert data_cen == pytest.approx(amplitude)
    assert output_dataset[0].err_hdr["LAYER_2"] == "fluxcal_error"
    #this is the error propagation of the multiplication with the flux calibration factor
    #and should agree with the value of the output image
    flux_err_cen = flux_dataset[0].err[0,512,512]
    err_layer2 = fluxcal_factor.fluxcal_err * el_cen
    err_comb = np.sqrt((flux_err_cen * fluxcal_factor.fluxcal_fac)**2 + err_layer2**2)
    assert output_dataset[0].err[0,512, 512] == pytest.approx(err_comb)
    assert output_dataset[0].err[1,512, 512] == pytest.approx(err_layer2)
    
    #Test the optional background subtraction#
    flux_image_back = create_flux_image(band_flux, fwhm, cal_factor, filter='3C', target_name='Vega', fsm_x=0.0, 
                      fsm_y=0.0, exptime=1.0, filedir=datadir, platescale=21.8, background=3,
                      add_gauss_noise=True, noise_scale=1., file_save=True)
    
    [flux_back, flux_err_back, back] = fluxcal.aper_phot(flux_image_back, radius, frac_enc_energy=0.997, method='subpixel', subpixels=5,
              background_sub=True, r_in=5, r_out=10, centering_method='xy', centroid_roi_radius=5)

    #calculated median background should be close to the input
    assert back == pytest.approx(3, abs = 0.03)
    #the found values should be close to the ones without background subtraction
    assert flux_back == pytest.approx(flux_el_ap, abs = 1)
    assert flux_err_back == pytest.approx(flux_err_ap, abs = 0.03)

    [flux_back, flux_err_back, back] = fluxcal.phot_by_gauss2d_fit(flux_image_back, fwhm, fit_shape=None, background_sub=True, r_in=5,
                        r_out=10, centering_method='xy', centroid_roi_radius=5)
    #calculated median background should be close to the input
    assert back == pytest.approx(3, abs = 0.03)
    #the found values should be close to the ones without background subtraction
    assert flux_back == pytest.approx(flux_el_gauss, abs = 1)
    assert flux_err_back == pytest.approx(flux_err_gauss, abs = 0.03)
    
    #Also test again the generation of the cal file now with a background subtraction
    aper_kwargs = {
        "encircled_radius": radius,
        "frac_enc_energy": 0.997,
        "method": "subpixel",
        "subpixels": 10,
        "background_sub": True,
        "r_in": 5,
        "r_out": 10,
        "centering_method": "xy",
        "centroid_roi_radius": 5
    }
    gauss_kwargs = {
        'fwhm': fwhm,                  
        'fit_shape': None,            
        'background_sub': True,        
        'r_in': 5.0,                   
        'r_out': 10.0,                
        'centering_method': 'xy',     
        'centroid_roi_radius': 5       
    }
    fluxcal_factor_back = fluxcal.calibrate_fluxcal_aper(flux_image_back, flux_or_irr = 'flux', phot_kwargs=aper_kwargs)
    assert fluxcal_factor_back.fluxcal_fac == pytest.approx(fluxcal_factor.fluxcal_fac)
    assert fluxcal_factor_back.ext_hdr["LOCBACK"] == back
    fluxcal_factor_back_gauss = fluxcal.calibrate_fluxcal_gauss2d(flux_image_back, flux_or_irr = 'flux', phot_kwargs=gauss_kwargs)
    assert fluxcal_factor_back_gauss.fluxcal_fac == pytest.approx(fluxcal_factor_gauss.fluxcal_fac)
    assert fluxcal_factor_back_gauss.ext_hdr["LOCBACK"] == back

    # test l4_to_tda.determine_flux
    input_dataset = Dataset([flux_image_back, flux_image_back])
    output_dataset = l4_to_tda.determine_flux(input_dataset, fluxcal_factor_back,  photo = "aperture", phot_kwargs = aper_kwargs)
    assert output_dataset[0].ext_hdr["FLUX"] == pytest.approx(band_flux)
    assert output_dataset[0].ext_hdr["LOCBACK"] == pytest.approx(3, abs = 0.03)
    #sanity check: vega is input source, so app mag 0
    assert output_dataset[0].ext_hdr["APP_MAG"] == pytest.approx(0.0)
    
    output_dataset = l4_to_tda.determine_flux(input_dataset, fluxcal_factor_back_gauss,  photo = "2dgauss", phot_kwargs = gauss_kwargs)
    assert output_dataset[0].ext_hdr["FLUX"] == pytest.approx(band_flux)
    assert output_dataset[0].ext_hdr["LOCBACK"] == pytest.approx(3, abs = 0.03)
    #sanity check: Vega is input source, so app mag 0
    assert output_dataset[0].ext_hdr["APP_MAG"] == pytest.approx(0.0)
    
    #estimate of the error propagated to the final flux
    flux_err_ap = np.sqrt(error_sum**2 * fluxcal_factor.fluxcal_fac**2 + fluxcal_factor.fluxcal_err**2 * 200**2)
    flux_err_gauss = np.sqrt(error_gauss**2 * fluxcal_factor_gauss.fluxcal_fac**2 + fluxcal_factor_gauss.fluxcal_err**2 * 200**2)
    
    input_dataset = Dataset([flux_image, flux_image])
    output_dataset = l4_to_tda.determine_flux(input_dataset, fluxcal_factor,  photo = "aperture", phot_kwargs = None)
    assert output_dataset[0].ext_hdr["FLUX"] == pytest.approx(band_flux)
    assert output_dataset[0].ext_hdr["FLUXERR"] == pytest.approx(flux_err_ap, rel = 0.1)
    assert output_dataset[0].ext_hdr["LOCBACK"] == 0
    mag_err_ap = 2.5/np.log(10) * flux_err_ap/band_flux
    
    assert output_dataset[0].ext_hdr["MAGERR"] == pytest.approx(mag_err_ap, rel = 0.1)
    
    output_dataset = l4_to_tda.determine_flux(input_dataset, fluxcal_factor_gauss,  photo = "2dgauss", phot_kwargs = None)
    assert output_dataset[0].ext_hdr["FLUX"] == pytest.approx(band_flux)
    assert output_dataset[0].ext_hdr["FLUXERR"] == pytest.approx(flux_err_gauss, rel = 0.1)
    assert output_dataset[0].ext_hdr["LOCBACK"] == 0
    mag_err_gauss = 2.5/np.log(10) * flux_err_gauss/band_flux
    
    assert output_dataset[0].ext_hdr["MAGERR"] == pytest.approx(mag_err_gauss, rel = 0.1)
    
    corgidrp.track_individual_errors = old_ind
    
if __name__ == '__main__':
    test_get_filter_name()
    test_flux_calc()
    test_colorcor()
    test_calspec_download()
    test_app_mag()
    test_fluxcal_file()
    test_abs_fluxcal()






from corgidrp.mocks import create_psfsub_dataset,create_default_headers
from corgidrp.l3_to_l4 import do_psf_subtraction
from corgidrp.data import PyKLIPDataset, Image, Dataset
from corgidrp.detector import nan_flags, flag_nans
from corgidrp.astrom import create_circular_mask
from scipy.ndimage import shift, rotate
import pytest
import numpy as np

## Helper functions/quantities

iwa_lod = 3.
owa_lod = 9.7
d = 2.36 #m
lam = 573.8e-9 #m
pixscale_arcsec = 0.0218

iwa_pix = iwa_lod * lam / d * 206265 / pixscale_arcsec
owa_pix = owa_lod * lam / d * 206265 / pixscale_arcsec

st_amp = 100.
noise_amp=1e-11
pl_contrast=1e-4
rel_tolerance = 0.05

## pyKLIP data class tests

def test_pyklipdata_ADI():
    """Tests that pyklip dataset centers frame, assigns rolls, and initializes PSF library properly for ADI data. 
    """

    rolls = [0,90]
    # Init with center shifted by 1 pixel in x, 2 pixels in y
    mock_sci,mock_ref = create_psfsub_dataset(2,0,rolls,
                                              centerxy=(50.5,51.5))

    pyklip_dataset = PyKLIPDataset(mock_sci,psflib_dataset=mock_ref)

    # Check image is centered properly
    for i,image in enumerate(pyklip_dataset._input):

        assert mock_sci.all_data[i,2:,1:] == pytest.approx(image[:-2,:-1]), f"Frame {i} centered improperly."

    # Check roll assignments and filenames match up for sci dataset
    for r,roll in enumerate(pyklip_dataset._PAs):
        assert roll == rolls[r]
        assert pyklip_dataset._filenames[r] == f'MOCK_sci_roll{roll}.fits_INT1', f"Incorrect roll assignment for frame {r}."
    
    # Check ref library is None
    assert pyklip_dataset.psflib is None, "pyklip_dataset.psflib is not None, even though no reference dataset was provided."

def test_pyklipdata_RDI():
    """Tests that pyklip dataset centers frame, assigns rolls, and initializes PSF library properly for RDI data. 
    """
    rolls = [45,180]
    n_sci = 1
    n_ref = 1
    # Init with center shifted
    mock_sci,mock_ref = create_psfsub_dataset(n_sci,n_ref,rolls,centerxy=(50.5,51.5))

    pyklip_dataset = PyKLIPDataset(mock_sci,psflib_dataset=mock_ref)
    
    # Check image is centered properly
    for i,image in enumerate(pyklip_dataset._input):

        assert mock_sci.all_data[i,2:,1:] == pytest.approx(image[:-2,:-1]), f"Frame {i} centered improperly."

    # Check roll assignments and filenames match up for sci dataset
    for r,roll in enumerate(pyklip_dataset._PAs):
        assert roll == rolls[r]
        assert pyklip_dataset._filenames[r] == f'MOCK_sci_roll{roll}.fits_INT1', f"Incorrect roll assignment for frame {r}."
    
    # Check ref library shape
    assert pyklip_dataset._psflib.master_library.shape[0] == n_sci+n_ref
 
def test_pyklipdata_ADIRDI():
    """Tests that pyklip dataset centers frame, assigns rolls, and initializes PSF library properly for ADI+RDI data. 
    """
    rolls = [45,-45,180]
    n_sci = 2
    n_ref = 1
    # Init with center shifted by 1 pixel
    mock_sci,mock_ref = create_psfsub_dataset(n_sci,n_ref,rolls,
                                              centerxy=(50.5,51.5))

    pyklip_dataset = PyKLIPDataset(mock_sci,psflib_dataset=mock_ref)

    # Check image is recentered properly
    for i,image in enumerate(pyklip_dataset._input):

        assert mock_sci.all_data[i,2:,1:] == pytest.approx(image[:-2,:-1]), f"Frame {i} centered improperly."

    # Check roll assignments and filenames match up for sci dataset
    for r,roll in enumerate(pyklip_dataset._PAs):
        assert roll == rolls[r]
        assert pyklip_dataset._filenames[r] == f'MOCK_sci_roll{roll}.fits_INT1', f"Incorrect roll assignment for frame {r}."
    
    # Check ref library shape
    assert pyklip_dataset._psflib.master_library.shape[0] == n_sci+n_ref

def test_pyklipdata_badtelescope():
    """Tests that pyklip data class initialization fails if data does not come from Roman.
    """
    mock_sci,mock_ref = create_psfsub_dataset(1,1,[0,0])
    mock_sci[0].pri_hdr['TELESCOP'] = "HUBBLE"

    with pytest.raises(UserWarning):
        _ = PyKLIPDataset(mock_sci,psflib_dataset=mock_ref)

def test_pyklipdata_badinstrument():
    """Tests that pyklip data class initialization fails if data does not come from Coronagraph Instrument.
    """
    mock_sci,mock_ref = create_psfsub_dataset(1,1,[0,0])
    mock_sci[0].pri_hdr['INSTRUME'] = "WFI"

    with pytest.raises(UserWarning):
        _ = PyKLIPDataset(mock_sci,psflib_dataset=mock_ref)

def test_pyklipdata_badcfamname():
    """Tests that pyklip data class raises an error if the CFAM position is not a valid position name.
    """
    mock_sci,mock_ref = create_psfsub_dataset(1,1,[0,0])
    mock_sci[0].ext_hdr['CFAMNAME'] = "BAD"

    with pytest.raises(UserWarning):
        _ = PyKLIPDataset(mock_sci,psflib_dataset=mock_ref)

def test_pyklipdata_notdataset():
    """Tests that pyklip data class raises an error if the iput is not a corgidrp dataset object.
    """
    mock_sci,mock_ref = create_psfsub_dataset(1,0,[0])
    mock_ref = []
    with pytest.raises(UserWarning):
        _ = PyKLIPDataset(mock_sci,psflib_dataset=mock_ref)

    mock_sci = []
    mock_ref = None
    with pytest.raises(UserWarning):
        _ = PyKLIPDataset(mock_sci,psflib_dataset=mock_ref)

def test_pyklipdata_badimgshapes():
    """Tests that pyklip data class enforces all data frames to have the same shape.
    """
    mock_sci,mock_ref = create_psfsub_dataset(2,0,[0,0])
    
    mock_sci[0].data = np.zeros((5,5))
    with pytest.raises(UserWarning):
        _ = PyKLIPDataset(mock_sci,psflib_dataset=mock_ref)

def test_pyklipdata_multiplepixscales():
    """Tests that pyklip data class enforces that each frame has the same pixel scale.
    """
    mock_sci,mock_ref = create_psfsub_dataset(2,0,[0,0])
    
    mock_sci[0].ext_hdr["PLTSCALE"] = 10
    with pytest.raises(UserWarning):
        _ = PyKLIPDataset(mock_sci,psflib_dataset=mock_ref)

# DQ flagging tests

def make_test_data(frame_shape,n_frames=1,):
    """Makes a test corgidrp Dataset of all zeros with the desired
    frame shape and number of frames.

    Args:
        frame_shape (listlike of int): 2D or 3D image shape desired.
        n_frames (int, optional): Number of frames. Defaults to 1.

    Returns:
        corgidrp.Dataset: mock dataset of all zeros.
    """
    
    frames = []
    for i in range(n_frames):
        prihdr, exthdr = create_default_headers()
        im_data = np.zeros(frame_shape).astype(np.float64)
        frame = Image(im_data, pri_hdr=prihdr, ext_hdr=exthdr)

        frames.append(frame)

    dataset = Dataset(frames)
    return dataset

def test_nanflags_2D():
    """Test detector.nan_flags() on 2D data.
    """

    # 2D:
    mock_dataset = make_test_data([10,10],n_frames=2,)
    mock_dataset.all_dq[0,4,6] = 1
    mock_dataset.all_dq[1,5,7] = 1

    expected_data = np.zeros([2,10,10])
    expected_data[0,4,6] = np.nan
    expected_data[1,5,7] = np.nan

    expected_err = np.zeros([2,1,10,10])
    expected_err[0,0,4,6] = np.nan
    expected_err[1,0,5,7] = np.nan

    nanned_dataset = nan_flags(mock_dataset,threshold=1)

    if not np.array_equal(nanned_dataset.all_data, expected_data,equal_nan=True):
        # import matplotlib.pyplot as plt 
        # fig,axes = plt.subplots(1,2)
        # axes[0].imshow(nanned_dataset.all_data[0,:,:])
        # axes[1].imshow(expected_data[0,:,:])
        # plt.show()
        # plt.close()
        raise Exception('2D nan_flags test produced unexpected result')

    if not np.array_equal(nanned_dataset.all_err,expected_err,equal_nan=True):
        raise Exception('2D nan_flags test produced unexpected result for ERR array')

def test_nanflags_3D():
    """Test detector.nan_flags() on 3D data.
    """

    # 3D:
    mock_dataset = make_test_data([3,10,10],n_frames=2,)
    mock_dataset.all_dq[0,0,4,6] = 1
    mock_dataset.all_dq[1,:,5,7] = 1

    expected_data = np.zeros([2,3,10,10])
    expected_data[0,0,4,6] = np.nan
    expected_data[1,:,5,7] = np.nan

    expected_err = np.zeros([2,1,3,10,10])
    expected_err[0,0,0,4,6] = np.nan
    expected_err[1,0,:,5,7] =  np.nan

    nanned_dataset = nan_flags(mock_dataset,threshold=1)

    if not np.array_equal(nanned_dataset.all_data, expected_data,equal_nan=True):
        raise Exception('2D nan_flags test produced unexpected result')
    
    if not np.array_equal(nanned_dataset.all_err, expected_err,equal_nan=True):
        raise Exception('3D nan_flags test produced unexpected result for ERR array')

def test_nanflags_mixed_dqvals():
    """Test detector.nan_flags() on 3D data with some DQ values below the threshold.
    """

    # 3D:
    mock_dataset = make_test_data([3,10,10],n_frames=2,)
    mock_dataset.all_dq[0,0,4,6] = 1
    mock_dataset.all_dq[1,:,5,7] = 2

    expected_data = np.zeros([2,3,10,10])
    expected_data[1,:,5,7] = np.nan

    expected_err = np.zeros([2,1,3,10,10])
    expected_err[1,0,:,5,7] =  np.nan

    nanned_dataset = nan_flags(mock_dataset,threshold=2)

    if not np.array_equal(nanned_dataset.all_data, expected_data,equal_nan=True):
        raise Exception('nan_flags with mixed dq values produced unexpected result')
    
    if not np.array_equal(nanned_dataset.all_err, expected_err,equal_nan=True):
        raise Exception('3D nan_flags with mixed dq values produced unexpected result for ERR array')

def test_flagnans_2D():
    """Test detector.flag_nans() on 2D data.
    """

    # 2D:
    mock_dataset = make_test_data([10,10],n_frames=2,)
    mock_dataset.all_data[0,4,6] = np.nan
    mock_dataset.all_data[1,5,7] = np.nan

    expected_dq = np.zeros([2,10,10])
    expected_dq[0,4,6] = 1
    expected_dq[1,5,7] = 1

    flagged_dataset = flag_nans(mock_dataset)

    if not np.array_equal(flagged_dataset.all_dq, expected_dq,equal_nan=True):
        raise Exception('2D nan_flags test produced unexpected result for DQ array')
    
def test_flagnans_3D():
    """Test detector.flag_nans() on 3D data.
    """

    # 3D:
    mock_dataset = make_test_data([3,10,10],n_frames=2,)
    mock_dataset.all_data[0,0,4,6] = np.nan
    mock_dataset.all_data[1,:,5,7] = np.nan

    expected_dq = np.zeros([2,3,10,10])
    expected_dq[0,0,4,6] = 1
    expected_dq[1,:,5,7] = 1
    
    flagged_dataset = flag_nans(mock_dataset)

    if not np.array_equal(flagged_dataset.all_dq, expected_dq,equal_nan=True):
        raise Exception('3D flag_nans test produced unexpected result for DQ array')

def test_flagnans_flagval2():
    """Test detector.flag_nans() on 3D data with a non-default DQ value.
    """

    # 3D:
    mock_dataset = make_test_data([3,10,10],n_frames=2,)
    mock_dataset.all_data[0,0,4,6] = np.nan
    mock_dataset.all_data[1,:,5,7] = np.nan

    expected_dq = np.zeros([2,3,10,10])
    expected_dq[0,0,4,6] = 2
    expected_dq[1,:,5,7] = 2

    flagged_dataset = flag_nans(mock_dataset,flag_val=2)

    if not np.array_equal(flagged_dataset.all_dq, expected_dq,equal_nan=True):
        raise Exception('3D nan_flags test produced unexpected result for DQ array')

## PSF subtraction step tests

def test_psf_sub_split_dataset():
    """Tests that psf subtraction step can correctly split an input dataset into
    science and reference dataset, if they are not passed in separately.
    """

    # Sci & Ref
    numbasis = [1,4,8]
    rolls = [270+13,270-13,0,0]
    mock_sci,mock_ref = create_psfsub_dataset(2,2,rolls,
                                              st_amp=st_amp,
                                              noise_amp=noise_amp,
                                              pl_contrast=pl_contrast)
    
    # combine mock_sci and mock_ref into 1 dataset
    frames = [*mock_sci,*mock_ref]
    mock_sci_and_ref = Dataset(frames)

    # Pass combined dataset to do_psf_subtraction
    result = do_psf_subtraction(mock_sci_and_ref,
                                numbasis=numbasis,
                                fileprefix='test_single_dataset',
                                do_crop=False,
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False)
    
    # Should choose ADI+RDI
    for frame in result:
        if not frame.pri_hdr['KLIP_ALG'] == 'ADI+RDI':
            raise Exception(f"Chose {frame.pri_hdr['KLIP_ALG']} instead of 'ADI+RDI' mode when provided 2 science images and 2 references.")

    # Try passing only science frames
    result = do_psf_subtraction(mock_sci,
                                numbasis=numbasis,
                                fileprefix='test_sci_only_dataset',
                                do_crop=False,
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False)
    
    # Should choose ADI
    for frame in result:
        if not frame.pri_hdr['KLIP_ALG'] == 'ADI':
            raise Exception(f"Chose {frame.pri_hdr['KLIP_ALG']} instead of 'ADI' mode when provided 2 science images and no references.")

    # pass only reference frames (should fail)
    with pytest.raises(UserWarning):
        _ = do_psf_subtraction(mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_ref_only_dataset',
                                do_crop=False,
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False)

def test_psf_sub_ADI_nocrop():
    """Tests that psf subtraction step correctly identifies an ADI dataset (multiple rolls, no references), 
    that overall counts decrease, that the KLIP result matches the analytical expectation, and that the 
    output data shape is correct.
    """

    numbasis = [1]
    rolls = [270+13,270-13]
    mock_sci,mock_ref = create_psfsub_dataset(2,0,rolls,
                                              st_amp=st_amp,
                                              noise_amp=noise_amp,
                                              pl_contrast=pl_contrast)

    result = do_psf_subtraction(mock_sci,reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_ADI',
                                do_crop=False,
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False)

    analytical_result = shift((rotate(mock_sci[0].data - mock_sci[1].data,-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - mock_sci[0].data,-rolls[1],reshape=False,cval=0)) / 2,
                              [0.5,0.5],
                              cval=np.nan)
    
    frame = result[0]
    for i,img in enumerate(frame.data):

        # import matplotlib.pyplot as plt

        # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
        # im0 = axes[0].imshow(img,origin='lower')
        # plt.colorbar(im0,ax=axes[0],shrink=0.8)
        # axes[0].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'])
        # axes[0].set_title(f'PSF Sub Result ({numbasis[i]} KL Modes)')

        # im1 = axes[1].imshow(analytical_result,origin='lower')
        # plt.colorbar(im1,ax=axes[1],shrink=0.8)
        # axes[1].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'])
        # axes[1].set_title('Analytical result')

        # diff = img - analytical_result
        # im2 = axes[2].imshow(diff,origin='lower')
        # plt.colorbar(im2,ax=axes[2],shrink=0.8)
        # axes[2].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'])
        # axes[2].set_title('Difference')

        # fig.suptitle('ADI')

        # plt.show()
        # plt.close()
    
        # Overall counts should decrease        
        if not np.nansum(mock_sci[0].data) > np.nansum(frame.data):
            raise Exception(f"ADI subtraction resulted in increased counts for frame {i}.")
                
        # Result should match analytical result for first KL mode       
        if np.nanmax(np.abs(frame.data[0] - analytical_result)) > np.nanmax(analytical_result) * rel_tolerance:
            raise Exception(f"Relative difference between ADI result and analytical result is greater then 5%.")
        
        if not frame.pri_hdr['KLIP_ALG'] == 'ADI':
            raise Exception(f"Chose {frame.pri_hdr['KLIP_ALG']} instead of 'ADI' mode when provided 2 science images and no references.")

    # Check expected data shape
    expected_data_shape = (1,len(numbasis),*mock_sci[0].data.shape)
    if not result.all_data.shape == expected_data_shape:
        raise Exception(f"Result data shape was {result.all_data.shape} instead of expected {expected_data_shape} after ADI subtraction.")

def test_psf_sub_RDI_nocrop(): 
    """Tests that psf subtraction step correctly identifies an RDI dataset (single roll, 1 or more references), 
    that overall counts decrease, that the KLIP result matches the analytical expectation, and that the 
    output data shape is correct.
    """
    numbasis = [1]
    rolls = [13,0]

    mock_sci,mock_ref = create_psfsub_dataset(1,1,rolls,ref_psf_spread=1.,
                                centerxy=(49.5,49.5),
                                pl_contrast=pl_contrast,
                                noise_amp=noise_amp,
                                st_amp=st_amp
                                )

    result = do_psf_subtraction(mock_sci,
                                reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_RDI',
                                do_crop=False,
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False
                                )
    analytical_result = rotate(mock_sci[0].data - mock_ref[0].data,-rolls[0],reshape=False,cval=np.nan)
    
    for i,frame in enumerate(result):

        mask = create_circular_mask(frame.data.shape[-2:],r=iwa_pix,center=(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY']))
        masked_frame = np.where(mask,np.nan,frame.data)

        # import matplotlib.pyplot as plt

        # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
        # im0 = axes[0].imshow(mock_sci[0].data,origin='lower')
        # plt.colorbar(im0,ax=axes[0],shrink=0.8)
        # axes[0].scatter(mock_sci[0].ext_hdr['STARLOCX'],mock_sci[0].ext_hdr['STARLOCY'])
        # axes[0].set_title(f'Sci Input')

        # im1 = axes[1].imshow(mock_ref[0].data,origin='lower')
        # plt.colorbar(im1,ax=axes[1],shrink=0.8)
        # axes[1].scatter(mock_ref[0].ext_hdr['STARLOCX'],mock_ref[0].ext_hdr['STARLOCY'])
        # axes[1].set_title('Ref Input')

        # im2 = axes[2].imshow(mock_sci[0].data - mock_ref[0].data,origin='lower')
        # plt.colorbar(im2,ax=axes[2],shrink=0.8)
        # axes[2].scatter(mock_sci[0].ext_hdr['STARLOCX'],mock_sci[0].ext_hdr['STARLOCY'])
        # axes[2].set_title('Difference')

        # fig.suptitle('Inputs')

        # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
        # im0 = axes[0].imshow(frame.data - np.nanmedian(frame.data),origin='lower')
        # plt.colorbar(im0,ax=axes[0],shrink=0.8)
        # axes[0].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'])
        # axes[0].set_title(f'PSF Sub Result ({numbasis[i]} KL Modes, Median Subtracted)')

        # im1 = axes[1].imshow(analytical_result,origin='lower')
        # plt.colorbar(im1,ax=axes[1],shrink=0.8)
        # axes[1].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'])
        # axes[1].set_title('Analytical result')

        # norm = LogNorm(vmin=1e-8, vmax=1, clip=False)
        # im2 = axes[2].imshow(frame.data - np.nanmedian(frame.data) - analytical_result,
        #                      origin='lower',norm=None)
        # plt.colorbar(im2,ax=axes[2],shrink=0.8)
        # axes[2].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'])
        # axes[2].set_title('Difference')

        # fig.suptitle('RDI Result')

        # plt.show()
        # plt.close()
        
        # Overall counts should decrease        
        if not np.nansum(mock_sci[0].data) > np.nansum(frame.data):
            raise Exception(f"RDI subtraction resulted in increased counts for frame {i}.")
        
        # The step should choose mode RDI based on having 1 roll and 1 reference.
        if not frame.pri_hdr['KLIP_ALG'] == 'RDI':
            raise Exception(f"Chose {frame.pri_hdr['KLIP_ALG']} instead of 'RDI' mode when provided 1 science image and 1 reference.")
        
        # Frame should match analytical result outside of the IWA (after correcting for the median offset)
        if not np.nanmax(np.abs((masked_frame - np.nanmedian(frame.data)) - analytical_result)) < 1e-5:
            raise Exception("RDI subtraction did not produce expected analytical result.")
    
    # Check expected data shape
    expected_data_shape = (1,len(numbasis),*mock_sci[0].data.shape)
    if not result.all_data.shape == expected_data_shape:
        raise Exception(f"Result data shape was {result.all_data.shape} instead of expected {expected_data_shape} after RDI subtraction.")
    
def test_psf_sub_ADIRDI_nocrop():
    """Tests that psf subtraction step correctly identifies an ADI+RDI dataset (multiple rolls, 1 or more references), 
    that overall counts decrease, that the KLIP result matches the analytical expectation for 1 KL mode, and that the 
    output data shape is correct.
    """

    numbasis = [1,2,3,4]
    rolls = [13,-13,0]
    mock_sci,mock_ref = create_psfsub_dataset(2,1,rolls,
                                              st_amp=st_amp,
                                              noise_amp=noise_amp,
                                              pl_contrast=pl_contrast)
    

    analytical_result1 = (rotate(mock_sci[0].data - (mock_sci[1].data/2+mock_ref[0].data/2),-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - (mock_sci[0].data/2+mock_ref[0].data/2),-rolls[1],reshape=False,cval=0)) / 2
    analytical_result2 = (rotate(mock_sci[0].data - mock_sci[1].data,-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - mock_sci[0].data,-rolls[1],reshape=False,cval=0)) / 2                         
    analytical_results = [analytical_result1,analytical_result2]
    
    result = do_psf_subtraction(mock_sci,reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_ADI+RDI',
                                do_crop=False,
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False)
    
    for i,frame in enumerate(result):

        
        mask = create_circular_mask(frame.data.shape[-2:],r=iwa_pix,center=(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY']))
        masked_frame = np.where(mask,np.nan,frame.data)

        # import matplotlib.pyplot as plt

        # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
        # im0 = axes[0].imshow(frame.data - np.nanmedian(frame.data),origin='lower')
        # plt.colorbar(im0,ax=axes[0],shrink=0.8)
        # axes[0].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'])
        # axes[0].set_title(f'PSF Sub Result ({numbasis[i]} KL Modes, Median Subtracted)')

        # im1 = axes[1].imshow(analytical_results[0],origin='lower')
        # plt.colorbar(im1,ax=axes[1],shrink=0.8)
        # axes[1].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'])
        # axes[1].set_title('Analytical result')

        # im2 = axes[2].imshow(masked_frame - np.nanmedian(frame.data) - analytical_results[0],origin='lower')
        # plt.colorbar(im2,ax=axes[2],shrink=0.8)
        # axes[2].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'])
        # axes[2].set_title('Difference')

        # fig.suptitle('ADI+RDI')

        # plt.show()
        # plt.close()

        # Overall counts should decrease        
        if not np.nansum(mock_sci[0].data) > np.nansum(frame.data):
            raise Exception(f"ADI+RDI subtraction resulted in increased counts for frame {i}.")
        
        # Corgidrp should know to choose ADI+RDI mode
        if not frame.pri_hdr['KLIP_ALG'] == 'ADI+RDI':
            raise Exception(f"Chose {frame.pri_hdr['KLIP_ALG']} instead of 'ADI+RDI' mode when provided 2 science images and 1 reference.")
        
        # Frame should match analytical result outside of the IWA (after correcting for the median offset) for KL mode 1
        if i==0:
            if not np.nanmax(np.abs((masked_frame - np.nanmedian(frame.data)) - analytical_results[i])) < 1e-5:
                raise Exception("ADI+RDI subtraction did not produce expected analytical result.")
                
    # Check expected data shape
    expected_data_shape = (1,len(numbasis),*mock_sci[0].data.shape)
    if not result.all_data.shape == expected_data_shape:
        raise Exception(f"Result data shape was {result.all_data.shape} instead of expected {expected_data_shape} after ADI+RDI subtraction.")

def test_psf_sub_withcrop():
    """Tests that psf subtraction step results in the correct data shape when 
    cropping by default, and that overall counts decrease.
    """

    numbasis = [1,2]
    rolls = [270+13,270-13]
    mock_sci,mock_ref = create_psfsub_dataset(2,0,rolls,pl_contrast=1e-3)

    result = do_psf_subtraction(mock_sci,reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_withcrop',
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False)

    for i,frame in enumerate(result):
    
        # Overall counts should decrease        
        if not np.nansum(mock_sci[0].data) > np.nansum(frame.data):
            raise Exception(f"PSF subtraction resulted in increased counts for frame {i}.")

    # Check expected data shape
    expected_data_shape = (1,len(numbasis),60,60)
    if not result.all_data.shape == expected_data_shape:
        raise Exception(f"Result data shape was {result.all_data.shape} instead of expected {expected_data_shape} after ADI subtraction.")

def test_psf_sub_badmode():
    """Tests that psf subtraction step fails correctly if an unconfigured mode is supplied (e.g. SDI).
    """

    numbasis = [1,2,3,4]
    rolls = [13,-13,0]
    mock_sci,mock_ref = create_psfsub_dataset(2,1,rolls,
                                              st_amp=st_amp,
                                              noise_amp=noise_amp,
                                              pl_contrast=pl_contrast)
    

    with pytest.raises(Exception):
        _ = do_psf_subtraction(mock_sci,reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                mode='SDI',
                                fileprefix='test_SDI',
                                do_crop=False,
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False)
    
if __name__ == '__main__':  
    # test_pyklipdata_ADI()
    # test_pyklipdata_RDI()
    # test_pyklipdata_ADIRDI()
    # test_pyklipdata_badtelescope()
    # test_pyklipdata_badinstrument()
    # test_pyklipdata_badcfamname()
    # test_pyklipdata_notdataset()
    # test_pyklipdata_badimgshapes()
    # test_pyklipdata_multiplepixscales()

    # test_nanflags_2D()
    # test_nanflags_3D() 
    # test_nanflags_mixed_dqvals()
    # test_flagnans_2D()
    # test_flagnans_3D()
    # test_flagnans_flagval2()

    #test_psf_sub_split_dataset()

    # test_psf_sub_ADI_nocrop()
    test_psf_sub_RDI_nocrop()
    test_psf_sub_ADIRDI_nocrop()
    # test_psf_sub_withcrop()
    # test_psf_sub_badmode()



"""
Unit tests for check.py.
For each check, test with valid and invalid inputs for all three inputs.

Test valid here as well since most other functions rely on these for
error checking
"""
import pytest

import numpy as np

from corgidrp import check


# Invalid values

# string
strlist = [1j, None, (1.,), [5, 5], -1, 0, 1.0]
# real scalar
rslist = [1j, None, (1.,), [5, 5], 'txt']
# real nonnegative scalar
rnslist = [1j, None, (1.,), [5, 5], 'txt', -1]
# real positive scalar
rpslist = [1j, None, (1.,), [5, 5], 'txt', -1, 0]
# real scalar integer
rsilist = [1j, None, (1.,), [5, 5], 'txt', 1.0]
# nonnegative scalar integer
nsilist = [1j, None, (1.,), [5, 5], 'txt', -1, 1.0]
# positive scalar integer
psilist = [1j, None, (1.,), [5, 5], 'txt', -1, 0, 1.0]
# real array
rarraylist = [1j*np.ones((5, 4)), (1+1j)*np.ones((5, 5, 5)), 'foo',
              np.array([[1, 2], [3, 4], [5, 'a']])]
# 1D array
oneDlist = [np.ones((5, 4)), np.ones((5, 5, 5)), 'foo']
# 2D array
twoDlist = [np.ones((5,)), np.ones((5, 5, 5)), [], 'foo']
# 2D square array
twoDsquarelist = [np.ones((5,)), np.ones((5, 5, 5)), np.ones((5, 4)),
                  [], 'foo']
# 3D array
threeDlist = [np.ones((5,)), np.ones((5, 5)), np.ones((2, 2, 2, 2)), [], 'foo']


class TestCheckException(Exception):
    pass


# real_positive_scalar
def test_real_positive_scalar_good():
    """
    Verify checker works correctly for valid input.

    Type: real positive scalar
    """
    try: 
        check.real_positive_scalar(1, 'rps', TestCheckException)
    except check.CheckException:
        pytest.xfail('real_positive_scalar failed on valid input')
    pass

def test_real_positive_scalar_bad_var():
    """
    fail on invalid variable type.
    
    Type: real positive scalar
    """
    for v0 in rpslist:
        with pytest.raises(TestCheckException):
            check.real_positive_scalar(v0, 'rps', TestCheckException)
            pass
        pass
    pass

def test_real_positive_scalar_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.real_positive_scalar(1, (1,), TestCheckException)
        pass
    pass

def test_real_positive_scalar_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.real_positive_scalar(1, 'rps', 'TestCheckException')
        pass
    pass
# real_nonnegative_scalar
def test_real_nonnegative_scalar_good():
    """
    Verify checker works correctly for valid input.
    
    Type: real nonnegative scalar
    """
    try:
        check.real_nonnegative_scalar(0, 'rps', TestCheckException)
    except check.CheckException:
        pytest.xfail('real_nonnegative_scalar failed on valid input')
    pass

def test_real_nonnegative_scalar_bad_var():
    """
    fail on invalid variable type.
    
    Type: real nonnegative scalar
    """
    for v0 in rnslist:
        with pytest.raises(TestCheckException):
            check.real_nonnegative_scalar(v0, 'rps', TestCheckException)
            pass
        pass
    pass

def test_real_nonnegative_scalar_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.real_nonnegative_scalar(0, (1,), TestCheckException)
        pass
    pass

def test_real_nonnegative_scalar_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.real_nonnegative_scalar(0, 'rps', 'TestCheckException')
        pass
    pass

# real_array
def test_real_array_good():
    """
    Verify checker works correctly for valid input.
    
    Type: real array
    """
    try:
        check.real_array(np.ones((5, 5)), 'real', TestCheckException)
    except check.CheckException:
        pytest.xfail('real_array failed on valid input')
    pass

def test_real_array_bad_var():
    """
    fail on invalid variable type.
    
    Type: real array
    """
    for v0 in rarraylist:
        with pytest.raises(TestCheckException):
            check.real_array(v0, '1D', TestCheckException)
            pass
        pass
    pass

def test_real_array_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.real_array(np.ones((5, 5)), (1,), TestCheckException)
        pass
    pass

def test_real_array_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.real_array(np.ones((5, )), 'rps', 'TestCheckException')
        pass
    pass

# oneD_array
def test_oneD_array_good():
    """
    Verify checker works correctly for valid input.
    
    Type: 1D array
    """
    try:
        check.oneD_array(np.ones((5, )), '1D', TestCheckException)
    except check.CheckException:
        pytest.xfail('oneD_array failed on valid input')
    pass

def test_oneD_array_bad_var():
    """
    fail on invalid variable type.

    Type: 1D array
    """
    for v0 in oneDlist:
        with pytest.raises(TestCheckException):
            check.oneD_array(v0, '1D', TestCheckException)
            pass
        pass
    pass

def test_oneD_array_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.oneD_array(np.ones((5, )), (1,), TestCheckException)
        pass
    pass

def test_oneD_array_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.oneD_array(np.ones((5, )), 'rps', 'TestCheckException')
        pass
    pass

# twoD_array
def test_twoD_array_good():
    """
    Verify checker works correctly for valid input.
    
    Type: 2D array
    """
    try:
        check.twoD_array(np.ones((5, 5)), '2d', TestCheckException)
    except check.CheckException:
        pytest.xfail('twoD_array failed on valid input')
    pass

def test_twoD_array_bad_var():
    """
    fail on invalid variable type.
    
    Type: 2D array
    """
    for v0 in twoDlist:
        with pytest.raises(TestCheckException):
            check.twoD_array(v0, '2d', TestCheckException)
            pass
        pass
    pass

def test_twoD_array_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.twoD_array(np.ones((5, 5)), (1,), TestCheckException)
        pass
    pass

def test_twoD_array_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.twoD_array(np.ones((5, 5)), 'rps', 'TestCheckException')
        pass
    pass

# twoD_square_array
def test_twoD_square_array_good():
    """
    Verify checker works correctly for valid input.

    Type: 2D array
    """
    try:
        check.twoD_array(np.ones((5, 5)), '2d', TestCheckException)
    except check.CheckException:
        pytest.xfail('twoD_square_array failed on valid input')
    pass

def test_twoD_square_array_bad_var():
    """
    fail on invalid variable type.

    Type: 2D array
    """
    for v0 in twoDsquarelist:
        with pytest.raises(TestCheckException):
            check.twoD_square_array(v0, '2d', TestCheckException)
            pass
        pass
    pass

def test_twoD_square_array_bad_var_shape():
    """
    fail on invalid variable type.
    
    Type: 2D square array
    """
    for v0 in [np.ones((5, 4)), np.ones((4, 6))]:
        with pytest.raises(TestCheckException):
            check.twoD_square_array(v0, '2d', TestCheckException)
            pass
        pass
    pass

def test_twoD_square_array_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.twoD_square_array(np.ones((5, 5)), (1,), TestCheckException)
        pass
    pass

def test_twoD_square_array_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.twoD_square_array(np.ones((5, 5)), 'rps',
                            'TestCheckException')
        pass
    pass

# threeD_array
def test_threeD_array_good():
    """
    Verify checker works correctly for valid input.

    Type: 3D array
    """
    try:
        check.threeD_array(np.ones((5, 5, 2)), '3d', TestCheckException)
    except check.CheckException:
        pytest.xfail('threeD_array failed on valid input')
    pass

def test_threeD_array_bad_var():
    """
    fail on invalid variable type.

    Type: 3D array
    """
    for v0 in threeDlist:
        with pytest.raises(TestCheckException):
            check.threeD_array(v0, '3d', TestCheckException)
            pass
        pass
    pass

def test_threeD_array_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.threeD_array(np.ones((5, 5, 2)), (1,), TestCheckException)
        pass
    pass

def test_threeD_array_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.threeD_array(np.ones((5, 5, 2)), 'rps', 'TestCheckException')
        pass
    pass

# real_scalar
def test_real_scalar_good():
    """
    Verify checker works correctly for valid input.

    Type: real scalar
    """
    try:
        check.real_scalar(1, 'rs', TestCheckException)
    except check.CheckException:
        pytest.xfail('real_scalar failed on valid input')
    pass

def test_real_scalar_bad_var():
    """
    fail on invalid variable type.

    Type: real scalar
    """
    for v0 in rslist:
        with pytest.raises(TestCheckException):
            check.real_scalar(v0, 'rs', TestCheckException)
            pass
        pass
    pass

def test_real_scalar_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.real_scalar(1, (1,), TestCheckException)
        pass
    pass

def test_real_scalar_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.real_scalar(1, 'rs', 'TestCheckException')
        pass
    pass

# positive_scalar_integer
def test_positive_scalar_integer_good():
    """
    Verify checker works correctly for valid input.

    Type: positive scalar integer
    """
    try:
        check.positive_scalar_integer(1, 'psi', TestCheckException)
    except check.CheckException:
        pytest.xfail('positive_scalar_integer failed on valid input')
    pass

def test_positive_scalar_integer_bad_var():
    """
    fail on invalid variable type.

    Type: positive scalar integer
    """
    for v0 in psilist:
        with pytest.raises(TestCheckException):
            check.positive_scalar_integer(v0, 'psi', TestCheckException)
            pass
        pass
    pass

def test_positive_scalar_integer_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.positive_scalar_integer(1, (1,), TestCheckException)
        pass
    pass

def test_positive_scalar_integer_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.positive_scalar_integer(1, 'psi', 'TestCheckException')
        pass
    pass

# nonnegative_scalar_integer
def test_nonnegative_scalar_integer_good():
    """
    Verify checker works correctly for valid input.

    Type: nonnegative scalar integer
    """
    for j in [0, 1, 2]:
        try:
            check.nonnegative_scalar_integer(j, 'nsi', TestCheckException)
        except check.CheckException:
            pytest.xfail('nonnegative_scalar_integer failed on valid input')
        pass
    pass

def test_nonnegative_scalar_integer_bad_var():
    """
    fail on invalid variable type.
    
    Type: nonnegative scalar integer
    """
    for v0 in nsilist:
        with pytest.raises(TestCheckException):
            check.nonnegative_scalar_integer(v0, 'nsi', TestCheckException)
            pass
        pass
    pass

def test_nonnegative_scalar_integer_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.nonnegative_scalar_integer(1, (1,), TestCheckException)
        pass
    pass

def test_nonnegative_scalar_integer_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.nonnegative_scalar_integer(1, 'nsi', 'TestCheckException')
        pass
    pass

# scalar_integer
def test_scalar_integer_good():
    """
    Verify checker works correctly for valid input.

    Type: scalar integer
    """
    for j in [-2, -1, 0, 1, 2]:
        try:
            check.scalar_integer(j, 'si', TestCheckException)
        except check.CheckException:
            pytest.xfail('scalar_integer failed on valid input')
        pass
    pass

def test_scalar_integer_bad_var():
    """
    fail on invalid variable type.

    Type: scalar integer
    """
    for v0 in rsilist:
        with pytest.raises(TestCheckException):
            check.scalar_integer(v0, 'si', TestCheckException)
            pass
        pass
    pass

def test_scalar_integer_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.scalar_integer(1, (1,), TestCheckException)
        pass
    pass

def test_scalar_integer_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.scalar_integer(1, 'si', 'TestCheckException')
        pass
    pass

def test_string_good():
    """
    Verify checker works correctly for valid input.

    Type: string
    """
    for j in ['a', '1', '.']:
        try:
            check.string(j, 'string', TestCheckException)
        except check.CheckException:
            pytest.xfail('string failed on valid input')
        pass
    pass

def test_string_bad_var():
    """
    fail on invalid variable type.

    Type: string
    """
    for v0 in strlist:
        with pytest.raises(TestCheckException):
            check.string(v0, 'string', TestCheckException)
            pass
        pass
    pass

def test_string_bad_vname():
    """fail on invalid input name for user output."""
    with pytest.raises(check.CheckException):
        check.string('a', ('a',), TestCheckException)
        pass
    pass

def test_string_bad_vexc():
    """fail on input vexc not an Exception."""
    with pytest.raises(check.CheckException):
        check.scalar_integer('a', 'string', 'TestCheckException')
        pass
    pass


if __name__ == '__main__':
    test_nonnegative_scalar_integer_bad_var()
    test_nonnegative_scalar_integer_bad_vexc()
    test_nonnegative_scalar_integer_bad_vname()
    test_nonnegative_scalar_integer_good()
    test_oneD_array_bad_var()
    test_oneD_array_bad_vexc()
    test_oneD_array_bad_vname()
    test_oneD_array_good()
    test_positive_scalar_integer_bad_var()
    test_positive_scalar_integer_bad_vexc()
    test_positive_scalar_integer_bad_vname()
    test_positive_scalar_integer_good()
    test_real_array_bad_var()
    test_real_array_bad_vexc()
    test_real_array_bad_vname()
    test_real_array_good()
    test_real_nonnegative_scalar_bad_var()
    test_real_nonnegative_scalar_bad_vexc()
    test_real_nonnegative_scalar_bad_vname()
    test_real_nonnegative_scalar_good()
    test_real_positive_scalar_bad_var()
    test_real_positive_scalar_bad_vexc()
    test_real_positive_scalar_bad_vname()
    test_real_positive_scalar_good()
    test_real_scalar_bad_var()
    test_real_scalar_bad_vexc()
    test_real_scalar_bad_vname()
    test_real_scalar_good()
    test_scalar_integer_bad_var()
    test_scalar_integer_bad_vexc()
    test_scalar_integer_bad_vname()
    test_scalar_integer_good()
    test_string_bad_var()
    test_string_bad_vexc()
    test_string_bad_vname()
    test_string_good()
    test_threeD_array_bad_var()
    test_threeD_array_bad_vexc()
    test_threeD_array_bad_vname()
    test_threeD_array_good()
    test_twoD_array_bad_var()
    test_twoD_array_bad_vexc()
    test_twoD_array_bad_vname()
    test_twoD_array_good()
    test_twoD_square_array_bad_var()
    test_twoD_square_array_bad_var_shape()
    test_twoD_square_array_bad_vexc()
    test_twoD_square_array_bad_vname() 
    test_twoD_square_array_good()



import pytest
import os
import astropy.time as time
import corgidrp.mocks as mocks
import corgidrp
from astropy.io import fits
import numpy as np
import corgidrp.walker as walker
import corgidrp.caldb as caldb
import corgidrp.data as data
import corgidrp.detector as detector
from corgidrp.photon_counting import get_pc_mean, photon_count, PhotonCountException

def test_negative():
    """Values at or below the 
    threshold are photon-counted as 0, including negative values."""
    fr = np.array([-3,2,0])
    pc_fr = photon_count(fr, thresh=2)
    assert (pc_fr == np.array([0,0,0])).all()
    

def test_pc():
    '''
    Tests that a pixel that is masked heavily (reducing the number of usable frames for that pixel) has a bigger err than the average pixel.

    Tests that if masking is all the way through for a certain pixel, that pixel will 
    be flagged in the DQ. 
    
    Make sure there is failure when the number of iterations niter<1.

    Tests that the use of a mask to mask out bright pixels when photon-counting, to ensure failure for bright pixels outside region of interest when no mask used
    and success when mask used to mask those bright pixels.

    Tests safemode, which makes the function issue warnings instead of crashing (useful for HOWFSC's iterative digging of dark holes).

    Checks various inputs: threshold<0 gives exception, thresh>=em_gain gives exception, providing dataset of VISTYPE='DARK' while also inputting a photon-counted master dark raises exception, exception raised when 'EMGAIN_C' not the same for all frames.

    Checks various metadata changes: the expected filename and history edit for the output is done (for case of dark subtraction and the case of no dark subtraction), the master dark is indicated as such via a header keyword 'PC_STAT'.

    Tests that output average value of the photon-counted frames agrees with what was simulated (for the case of dark subtraction and no dark subtraction).

    Tests that if 'ISPC' is in image header and if it is False, get_pc_mean() raises an exception.

    Tests that an exception occurs if the user input "inputmode" is inconsistent with the input dataset type.

    Tests that an exception occurs if the photon-counted master dark does not have the header 'PCTHRESH'.

    Tests that an exception occurs if the photon-counted master dark has a 'PCTHRESH' value than the one to be used on the illuminated dataset.
    '''
    np.random.seed(555)
    dataset_err, dark_dataset_err, ill_mean, dark_mean = mocks.create_photon_countable_frames(Nbrights=160, Ndarks=160, cosmic_rate=0, full_frame=False, smear=False) 
    # instead of running through walker, just do the pre-processing steps simply
    # using EM gain=5000 and kgain=7 and bias=20000 and read noise = 100 and QE=0.9 (quantum efficiency), from mocks.create_photon_countable_frames()
    for f in dataset_err.frames:
        f.data = f.data.astype(float)*7 - 20000.
    for f in dark_dataset_err.frames:
        f.data = f.data.astype(float)*7 - 20000.
    dataset_err.all_data = dataset_err.all_data.astype(float)*7 - 20000.
    dark_dataset_err.all_data = dark_dataset_err.all_data.astype(float)*7 - 20000.
    # masked for half of the bright and half of the dark frames
    dataset_err.all_dq[80:,3,3] = 1
    dark_dataset_err.all_dq[80:,3,3] = 1
    dataset_err.all_dq[:, 2, 2] = 1 #masked all the way through for (2,2)
    for f in dataset_err.frames: 
        f.ext_hdr['RN'] = 100
        f.ext_hdr['KGAINPAR'] = 7
    for f in dark_dataset_err.frames: 
        f.ext_hdr['RN'] = 100
        f.ext_hdr['KGAINPAR'] = 7
    # process the frames to make PC dark
    pc_dark = get_pc_mean(dark_dataset_err, inputmode='darks')
    assert pc_dark.ext_hdr['PC_STAT'] == 'photon-counted master dark'
    # now process illuminated frames and subtract the PC dark
    pc_dataset_err = get_pc_mean(dataset_err, pc_master_dark=pc_dark)

    assert pc_dataset_err.frames[-1].filename == dataset_err[-1].filename

    history = ''
    for line in pc_dataset_err.frames[0].ext_hdr["HISTORY"]:
        history += line
    assert "Dark-subtracted: yes" in history
    assert np.isclose(pc_dataset_err.all_data.mean(), ill_mean - dark_mean, rtol=0.01) 
    # the DQ for that pixel should be 1
    assert pc_dataset_err[0].dq[2,2] == 1
    # err for (3,3) is above the 95th percentile of error: 
    assert pc_dataset_err[0].err[0][3,3]>np.nanpercentile(pc_dataset_err[0].err,95)

    # also when niter<1, exception
    with pytest.raises(PhotonCountException):
        get_pc_mean(dataset_err, niter=0)
    # when thresh<0, exception
    with pytest.raises(PhotonCountException):
        get_pc_mean(dataset_err, T_factor=-1, niter=2)
    # when thresh>=em_gain, exception
    with pytest.raises(Exception):
        get_pc_mean(dataset_err, T_factor=50, niter=2)
    # can't provide master dark while "VISTYPE" of input dataset is 'DARK'
    with pytest.raises(PhotonCountException):
        get_pc_mean(dark_dataset_err, pc_master_dark=pc_dark, inputmode='darks')
    # must have same 'CMDGAIN' and other header values throughout input dataset
    dark_dataset_err.frames[0].ext_hdr['EMGAIN_C'] = 4999
    with pytest.raises(PhotonCountException):
        get_pc_mean(dark_dataset_err, inputmode='dark')
    # change back:
    dark_dataset_err.frames[0].ext_hdr['EMGAIN_C'] = 5000

    # test to make sure PC dark's threshold matches the one used for illuminated frames 
    with pytest.raises(PhotonCountException):
        pc_dark.ext_hdr['PCTHRESH'] = 499
        get_pc_mean(dataset_err, pc_master_dark=pc_dark, inputmode='illuminated')
    # PC dark should have header 'PCTHRESH'
    with pytest.raises(PhotonCountException):
        pc_dark.ext_hdr.pop("PCTHRESH")
        get_pc_mean(dataset_err, pc_master_dark=pc_dark, inputmode='illuminated')
    # set it back:
    pc_dark.ext_hdr['PCTHRESH'] = 500
    # to trigger pc_ecount_max error
    for f in dataset_err.frames[:-2]: #all but 2 of the frames
        f.data[22:40,23:49] = np.abs(f.data.astype(float)[22:40,23:49]*3000)
    dataset_err.all_data[:-2,22:40,23:49] = np.abs(dataset_err.all_data.astype(float)[:-2,22:40,23:49]*3000)
    with pytest.raises(Exception):
        get_pc_mean(dataset_err, pc_ecount_max=1)
    # now use mask to exclude 22,23 from region of interest:
    mask = np.zeros_like(dataset_err.all_data[0])
    mask[22:40,23:49] = 1 #exclude
    thisfile_dir = os.path.dirname(__file__)
    mask_filepath = os.path.join(thisfile_dir, 'test_data', 'pc_mask.fits')
    hdr = fits.Header()
    prim = fits.PrimaryHDU(header=hdr)
    img = fits.ImageHDU(mask)
    hdul = fits.HDUList([prim,img])
    hdul.writeto(mask_filepath, overwrite=True)
    pc_dataset_err = get_pc_mean(dataset_err, pc_ecount_max=1, mask_filepath=mask_filepath)
    copy_pc = pc_dataset_err.all_data.copy()
    # mask out all the irrelevant pixels:
    copy_pc[0,22:40,23:49] = np.nan
    # didn't dark-subtract this time
    history = ''
    for line in pc_dataset_err.frames[0].ext_hdr["HISTORY"]:
        history += line
    assert "Dark-subtracted: no" in history
    assert np.isclose(np.nanmean(copy_pc), ill_mean, rtol=0.01) 
    # test safemode=False: should get warning instead of exception
    with pytest.warns(UserWarning):
        get_pc_mean(dataset_err, safemode=False)
    # test ISPC header value
    with pytest.raises(PhotonCountException):
        dataset_err.frames[0].ext_hdr['ISPC'] = False
        get_pc_mean(dataset_err, pc_master_dark=pc_dark)
    # set to True now
    dataset_err.frames[0].ext_hdr['ISPC'] = True
    # test inputmode's compatibility with dataset type
    with pytest.raises(PhotonCountException):
        get_pc_mean(dataset_err, pc_master_dark=pc_dark, inputmode='darks')
    with pytest.raises(PhotonCountException):
        get_pc_mean(dark_dataset_err, inputmode='illuminated')

def test_pc_subsets():
    '''
    Tests that the optional binning of frames works.
    '''
    np.random.seed(555)
    dataset_bin, dark_dataset_bin, ill_mean, dark_mean = mocks.create_photon_countable_frames(Nbrights=161, Ndarks=162, cosmic_rate=0, full_frame=False, smear=False) 
    # instead of running through walker, just do the pre-processing steps simply
    # using EM gain=5000 and kgain=7 and bias=20000 and read noise = 100 and QE=0.9 (quantum efficiency), from mocks.create_photon_countable_frames()
    for f in dataset_bin.frames:
        f.data = f.data.astype(float)*7 - 20000.
    for f in dark_dataset_bin.frames:
        f.data = f.data.astype(float)*7 - 20000.
    dataset_bin.all_data = dataset_bin.all_data.astype(float)*7 - 20000.
    dark_dataset_bin.all_data = dark_dataset_bin.all_data.astype(float)*7 - 20000.
    # process darks and check NUM_FR
    pc_dark = get_pc_mean(dark_dataset_bin, inputmode='darks', bin_size=40)
    assert pc_dark.ext_hdr['NUM_FR'] == 40 # The 2 remainder frames ignored for consistent statistics among the PC-averaged output frames
    assert 'Number of subsets: 4' in pc_dark.ext_hdr['HISTORY'][-4]
    # now process illuminated frames and subtract the PC dark
    pc_dataset = get_pc_mean(dataset_bin, pc_master_dark=pc_dark, bin_size=40)
    # bigger rtol below since we have fewer frames averaged over in each of the 161//40 = 162//40 = 4 frames in the output
    assert np.isclose(pc_dataset.all_data.mean(), ill_mean - dark_mean, rtol=0.06) 
    assert pc_dataset.frames[0].ext_hdr['NUM_FR'] == 40
    assert pc_dataset.frames[-1].ext_hdr['NUM_FR'] == 40 # The 1 remainder frame ignored for consistent statistics among the PC-averaged output frames
    assert 'Number of subsets: 4' in pc_dataset.frames[0].ext_hdr['HISTORY'][-2]
    with pytest.raises(PhotonCountException): #since number of frames in a dark subset would be less than that of a subset in illuminated
        get_pc_mean(dataset_bin, pc_master_dark=pc_dark, bin_size=51)
    # but this is fine:
    get_pc_mean(dataset_bin, pc_master_dark=pc_dark, bin_size=38)


if __name__ == '__main__':
    test_pc_subsets()
    test_pc()
    test_negative()
    
    